/**
 * @module-suite/bundle v0.6.7
 * Author: Gabe M <hey@gabe.mx>
 *
 * https://github.com/zelzen/module-suite
 */

'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var path = require('path');
var rollup = require('rollup');
var rollupPluginTerser = require('rollup-plugin-terser');
var url = _interopDefault(require('url'));
var got = _interopDefault(require('got'));
var rewriteModule = _interopDefault(require('@module-suite/rewrite'));

// Based on unpkg
const packageUrlFormat = /^\/?((?:@[^/@]+\/)?[^/@]+)(?:@([^/]+))?(\/.*)?$/;
function parsePackageURL(originalURL) {
  let {
    pathname,
    search,
    query
  } = url.parse(originalURL, true);
  if (pathname == null) return null;

  try {
    pathname = decodeURIComponent(pathname);
  } catch (error) {
    return null;
  }

  const match = packageUrlFormat.exec(pathname); // Disallow invalid URL formats.

  if (match == null) {
    return null;
  }

  let [_, packageName, packageVersion = 'latest', fileName = ''] = match; // Default to ".js" if non is specified

  if (fileName !== '' && path.extname(fileName) === '') {
    fileName += '.js';
  }

  return {
    // If the URL is /@scope/name@version/file.js?main=browser
    pathname,
    // /@scope/name@version/path.js
    search: search || '',
    // ?main=browser
    query,
    // { main: 'browser' }
    packageName,
    // @scope/name
    packageVersion,
    // version
    fileName // /file.js

  };
}

/**
 * Is ID referencing a local file path
 */
function isLocalIdentifier(id) {
  return id.startsWith('.') || id.startsWith('/');
}

function proxierImportPlugin(packageName, packageVersion, host, dependencies) {
  return {
    name: 'proxierImportPlugin',

    /* Resolve URL to download Source Code */
    resolveId(id, importer) {
      // console.log('id', id, importer);
      // If import id is a fully
      // qualified url, pass it along.
      if (id.startsWith('http')) {
        // If it's a self reference make
        // sure the non transpiled version is loaded in.
        if (id.includes(packageName)) return `${id}&transforms=false&minify=false`;
        return id;
      } // Local imports


      if (isLocalIdentifier(id)) {
        // console.log('resolve', id, importer);
        let importDir = ''; // Resolve root import dir for
        // modules loaded from the proxier.
        // This could include re-exports, and nested local imports.

        if (importer != null) {
          // Path import Dir from package if fully qualified url
          if (importer.startsWith(host)) {
            const importerModule = parsePackageURL(importer);

            if (importerModule !== null) {
              // Reference the same starting dir if a local path
              importDir = path.dirname(importerModule.fileName);
            }
          } // Else use referenced dir name
          else {
              importDir = path.dirname(importer);
            }
        } // If import dir is referencing the same
        // directory, set it to an empty string.


        importDir = importDir === '.' || importDir === '/' ? '' : `/${importDir}`; // console.log('id', id, importDir);

        return host + path.normalize(`/${packageName}@${packageVersion}${importDir}/${id}?output=esm&transforms=nodeenv,deadcode&minify=false`);
      } // External Imports


      return id;
    },

    /* Load Source Code */
    async load(id) {
      // Don't bundle imports for other packages
      if (id.includes(packageName) === false) return null; // Download imported module source

      const {
        body
      } = await got(id);
      return body;
    },

    /* Transform Loaded Source Code */
    transform(chunkCode, fileName) {
      // No need to transform if it's
      // already ran through rewrite-module on the server.
      if (fileName.startsWith('http')) return chunkCode;
      return rewriteModule(chunkCode, {
        host,
        packageName,
        packageVersion,
        currentFile: fileName,
        dependencies,
        // Minification will be ran on the full
        // bundle chunk from rollup.
        // No need to run per file.
        shouldMinify: false,
        // Output is always esm since
        // format will be handled by rollup.
        output: 'esm',
        // Replace NODE_ENV and remove deadcode
        transforms: new Set(['nodeenv', 'deadcode'])
      });
    }

  };
}

var name = "@module-suite/bundle";
var version = "0.6.7";

const banner = `/**
 * Bundled with ${name} v${version}
 * @preserve
 */`; // TODO: Support bundles with dynamic imports. Leave them as is.

async function bundleModule(code, {
  packageName,
  packageVersion,
  dependencies,
  host,
  shouldMinify = false,
  format = 'esm',
  rootFilePath
}) {
  console.log('[bundle-module] Bundling', packageName); // Resolves imported "file"
  // since we are loaded from memory.

  const importResolver = {
    name: 'importResolver',

    resolveId(id) {
      if (id === rootFilePath) return id;
    },

    load(id) {
      if (id === rootFilePath) return code;
      return null;
    }

  };
  const plugins = [importResolver, proxierImportPlugin(packageName, packageVersion, host, dependencies)]; // Minify full chunk with terser

  if (shouldMinify) plugins.push(rollupPluginTerser.terser());

  try {
    const bundle = await rollup.rollup({
      input: rootFilePath,
      plugins,

      external(id) {
        // console.log('external', id);
        // Self referenced imports over the CDN
        // should also be included.
        if (id.startsWith('http') && id.includes(packageName)) {
          return false;
        } // Locally referenced imports should be bundled.


        if (isLocalIdentifier(id)) return false; // Everything else is external
        // and shouldn't be bundled.

        return true;
      }

    }); // TODO: Limit output chunks to 1

    const {
      output
    } = await bundle.generate({
      format,
      banner,

      // Rewrite external imports
      // to be fully qualified urls.
      // e.g. "object-assign" => "https://module-cdn.exmaple.com/object-assign@^4.1.1?output=system"
      paths(id) {
        // Skip id which are already urls
        if (id.startsWith('http')) return id; // TODO: This doesn't rewrite dynamic imports
        // https://github.com/rollup/rollup/issues/2579

        const importedId = parsePackageURL(id); // console.log('paths', id, importedId);

        if (importedId === null) return id;
        const version = dependencies[importedId.packageName] || '*';

        if (version === '*') {
          console.warn(`[bundle-module] Unable to resolve package version for external dependency "${importedId.packageName}". Make sure to pass all dependencies.`);
        }

        return host + path.normalize(`/${importedId.packageName}@${version}${importedId.fileName}?output=${format}`);
      }

    });
    return output[0].code;
  } catch (err) {
    console.log('[bundle-module] An error occurred while bundling.');
    console.error(err); // Re-throw error so consume can catch

    throw err;
  }
}

module.exports = bundleModule;
